"""Test suite for the EventsQueue class."""
import os
import random

os.environ["CACHE_DIR"] = "tests/"
from dpypeline.akita.queue_events import EventsQueue


def delete_cache_file() -> None:
    """Delete the cache file containg the queue."""
    state_file = os.path.join(os.getenv("CACHE_DIR"), "queue_state.pickle")
    if os.path.isfile(state_file):
        os.remove(state_file)


def add_elements_to_queue(
    queue: EventsQueue, n: int = None
) -> tuple[EventsQueue, list[int]]:
    """
    Add elements to the queue.

    Parameters
    ----------
    queue
        Queue to add elements to.
    n, optional
        Number of elements to add, by default None
    """
    if n is None:
        n = random.randint(1, 1000)

    elements_to_add = [x for x in range(n)]
    for elem in elements_to_add:
        queue.enqueue(elem)

    return queue, elements_to_add


def test_queue_size() -> None:
    """Test the size of the queue."""
    delete_cache_file()

    # Create dummy queue
    queue = EventsQueue()

    # Add events to the queue
    n = random.randint(1, 1000)
    queue, _ = add_elements_to_queue(queue, n)

    assert queue.qsize() == n
    assert queue.get_queue_size() == n

    queue.clear_instance()


def test_enqueue_dequeue() -> None:
    """Test the enqueue method."""
    delete_cache_file()

    # Create dummy queue
    queue = EventsQueue()

    # Add events to the queue
    queue, added_elements = add_elements_to_queue(queue, 3)

    assert list(queue.queue) == added_elements

    # Remove two events from the queue
    queue.dequeue()
    queue.dequeue()

    assert list(queue.queue) == added_elements[2:]

    queue.clear_instance()


def test_peek() -> None:
    """Test the peek method."""
    delete_cache_file()

    # Create dummy queue
    queue = EventsQueue()

    # Peek when queue is Empty
    assert queue.peek() is None

    # Add events to the queue
    queue, added_elements = add_elements_to_queue(queue)

    # Peek the queue
    peeked_element = queue.peek()
    assert peeked_element == added_elements[0]
    assert list(queue.queue) == added_elements

    queue.clear_instance()


def test_singleton() -> None:
    """Test the singleton pattern."""
    delete_cache_file()

    # Create dummy queue
    queue = EventsQueue()

    # Add events to the queue
    _, _ = add_elements_to_queue(queue)

    queue_other = EventsQueue()

    assert queue is queue_other
    assert list(queue.queue) == list(queue_other.queue)

    queue.clear_instance()


def test_save_load_state() -> None:
    """Test the save and load of the state of the queue."""
    delete_cache_file()

    # Create dummy queue
    old_queue = EventsQueue()

    # Add events to the queue
    old_queue, added_elements = add_elements_to_queue(old_queue)

    # Destroy the queue
    old_queue.clear_instance()

    new_queue = EventsQueue()

    assert list(new_queue.queue) == added_elements

    # Add more events to the new queue
    new_queue, new_added_elements = add_elements_to_queue(new_queue)

    # Destroy the queue
    new_queue.clear_instance()

    new_new_queue = EventsQueue()

    assert list(new_new_queue.queue) == added_elements + new_added_elements

    new_new_queue.clear_instance()


def test_clear_instance() -> None:
    """Test the clear_instance method."""
    delete_cache_file()
    queue = EventsQueue()
    queue.clear_instance()
    queue_new = EventsQueue()

    assert queue is not queue_new



"""Thread-based ETL pipeline."""
import logging
from typing import Any

from .core import ETLPipeline, Job


class ThreadPipeline(ETLPipeline):
    """Thread-based ETL pipeline."""

    def __init__(self, jobs: list[Job] = None) -> None:
        """Initialize the ETL pipeline."""
        super().__init__(jobs)

    def produce_jobs(self, event: Any) -> Any:
        """
        Produce tasks triggered by an event.

        Parameters
        ----------
        event
            Triggering event.
        """
        for job in self._jobs:
            logging.info(f"Running job {job.name}.")
            logging.info(f"Running function {job.tasks[0].function.__name__}.")
            result = job.tasks[0].function(
                event, *job.tasks[0].args, **job.tasks[0].kwargs
            )
            for task in job.tasks[1:]:
                logging.info(f"Running function {task.function.__name__}.")
                result = task.function(result, *task.args, **task.kwargs)



"""S3 object store class."""
import json
import logging
import os

import fsspec
import s3fs


class ObjectStoreS3(s3fs.S3FileSystem):
    """
    S3 object store.

    Parameters
    ----------
    s3fs
        _description_
    """

    def __init__(
        self,
        anon: bool = False,
        store_credentials_json: str | None = None,
        secret: str | None = None,
        key: str | None = None,
        endpoint_url: str | None = None,
        *fs_args,
        **fs_kwargs,
    ) -> None:
        """
        Initialize the S3 object store.

        Parameters
        ----------
        anon, optional
            _description_, by default False
        store_credentials_json, optional
            _description_, by default None
        secret, optional
            _description_, by default None
        key, optional
            _description_, by default None
        endpoint_url, optional
            _description_, by default None
        """
        self._anon = anon
        logging.info("-" * 79)
        if store_credentials_json is None:
            logging.info(
                "No JSON file was provided."
                "Object store credentials will be obtained from the arguments passed."
            )
            self._store_credentials = {
                "secret": secret,
                "token": key,
                "endpoint_url": endpoint_url,
            }
        else:
            logging.info(
                f"Object store credentials will be read from the JSON file '{store_credentials_json}'"
            )
            self._store_credentials = self.load_store_credentials(
                store_credentials_json
            )

        self._remote_options = self.get_remote_options(override=True)

        super().__init__(*fs_args, **self._remote_options, **fs_kwargs)

    @staticmethod
    def load_store_credentials(path: str) -> dict:
        """
        Set the credentials of the object store from a JSON file.

        Parameters
        ----------
        path
            Absolute or relative filepath to the JSON file containing the object store credentials.

        Returns
        -------
        store_credentials
            Dictionary containing the values of the `token`, `secret` and `endpoint_url` keys used
            to access the object store.
        """
        try:
            with open(path) as f:
                store_credentials = json.load(f)
        except Exception as error:
            raise Exception(error)

        for key in ["token", "secret", "endpoint_url"]:
            if key not in store_credentials:
                logging.info("-" * 79)
                logging.warning(
                    f'"{key}" is not a key in the JSON file provided. Its value will be set to None.'
                )

        return store_credentials

    def get_remote_options(self, override: bool = False) -> dict:
        """
        Get the remote options of the object store.

        Parameters
        ----------
        override
            Flag to create remote_options from scratch (True) or to simply retrieve the current dict (False).

        Returns
        -------
        remote_options
            Dictionary containing the remote options of the object store.

        """
        if override:
            self._remote_options = {
                "anon": self._anon,
                "secret": self._store_credentials["secret"],
                "key": self._store_credentials["token"],
                "client_kwargs": {
                    "endpoint_url": self._store_credentials["endpoint_url"]
                },
            }

        return self._remote_options

    def get_mapper(
        self, bucket: str, prefix: str = "s3://", **get_mapper_kwargs
    ) -> fsspec.mapping.FSMap:
        """
        Make a MutableMaping interface to the desired bucket.

        Parameters
        ----------
        bucket
            Name of the bucket to place the file in.
        prefix: str, default "s3://"
            Protocol prefix
        **get_mapper_kwargs
            Kwargs for get_mapper. See: https://filesystem-spec.readthedocs.io/en/latest/api.html#fsspec.get_mapper.

        Returns
        -------
        mapper
            Dict-like key-value store.
        """
        mapper = fsspec.get_mapper(
            prefix + bucket, **self._remote_options, **get_mapper_kwargs
        )

        return mapper

    def get_bucket_list(self) -> list[str]:
        """
        Get the list of buckets in the object store.

        Returns
        -------
        bucket_list
            List of the object store buckets.
        """
        return self.ls("/")

    def write_file_to_bucket(self, path: str, bucket: str) -> None:
        """
        Write file from the local filesystem to a bucket of the object store.

        Parameters
        ----------
        path
            Absolute or relative filepath of the file to be written to the object store.
        bucket
            Name of the bucket to place the file in.
        """
        assert bucket in self.get_bucket_list(), f'Bucket "{bucket}" does not exist.'
        assert os.path.isfile(path), f'"{path}" is not a file.'

        with self.open(
            f"{bucket}/{path.rsplit('/', 1)[-1]}", mode="wb", s3=dict(profile="default")
        ) as fa:
            with open(path, mode="rb") as fb:
                fa.write(fb.read())



"""Celery Xarray tasks."""
import numpy as np
import xarray as xr
from celery.utils.log import get_task_logger

from ..app import app

logger = get_task_logger(__name__)


@app.task(bind=True)
def open_dataset(self, filepath: str, *args, **kwargs):
    """
    Open and decode a dataset.

    Notes
    -----
    https://docs.xarray.dev/en/stable/generated/xarray.open_dataset.html
    """
    logger.info(f"Opening xarray data set at {filepath}")
    return xr.open_dataset(filepath, *args, **kwargs)


@app.task(bind=True)
def to_zarr(self, dataset: xr.Dataset, *args, **kwargs):
    """
    Write dataset contents to a zarr group.

    Notes
    -----
    https://docs.xarray.dev/en/stable/generated/xarray.to_zarr.html
    """
    return dataset.to_zarr("trial.zarr", *args, **kwargs)


@app.task(bind=True)
def clean_dataset(
    self,
    dataset: xr.Dataset,
    fill_value=np.nan,
    supression_value: float = 1e20,
    threshold: float = 1e-6,
    *args,
    **kwargs,
) -> str:
    """Clean a dataset."""
    fill_value = fill_value if fill_value is not None else np.NaN
    supression_value = supression_value if supression_value is not None else np.NaN

    dataset_tmp = dataset.copy(deep=True).load()
    # dataset_tmp = dataset_tmp.fillna(fill_value)
    # for var in ["sos"]:#dataset.keys():
    # dataset_tmp[var].values = xr.where(np.abs(dataset_tmp[var] - supression_value) < threshold, fill_value, dataset_tmp[var])
    # dataset_tmp[var].encoding["_FillValue"] = fill_value

    cache_dir = "../"
    cached_file = f"{cache_dir}/transformed_{dataset.encoding['source'].rsplit('/', 1)[-1].rsplit('.', 1)[-2]}.nc"
    dataset_tmp.to_netcdf(cached_file)
    del dataset_tmp
    print("CACHED_FILE", cached_file)
    return cached_file



"""Celery app configuration."""
import os

assert (
    os.getenv("BROKER_URL") is not None
), "BROKER_URL environmental variable is not set."

broker_url = os.getenv("BROKER_URL")
# result_backend = "db+sqlite:///etl_pipeline.sqlite/"
imports = ("dpypeline.celery.tasks.xarray",)

task_serializer = "pickle"
result_serializer = "pickle"
accept_content = ["pickle"]

"""
task_routes = {
    "apds_file_mover.tasks.erddap_move": {"queue": "nrt_file_mover_erddap"},
    "apds_file_mover.tasks.archive_move": {"queue": "nrt_file_mover_archive"},
    "apds_file_mover.tasks.zip_files": {"queue": "nrt_zip_files"},
    "apds_file_mover.tasks.zip_and_move_file_to_erddap": {"queue": "nrt_file_zipper_mover"},
}

task_queues = {
    "nrt_file_mover_erddap": {"exchange": "data.outgoing", "routing_key": "NRT.file_mover.erddap"},
    "nrt_file_mover_archive": {"exchange": "data.outgoing", "routing_key": "NRT.file_mover.archive"},
    "nrt_zip_files": {"exchange": "data.outgoing", "routing_key": "NRT.zip_files"},
    "nrt_file_zipper_mover": {"exchange": "data.outgoing", "routing_key": "NRT.file.zipper_mover"},
}
"""

# Use extended results.
result_extended = True

# use custom table names for the database result backend.
# database_table_names = {"task": "etl_pipeline"}



"""Celer-based ETL pipeline."""
import logging
from typing import Any

from celery import chain, group

from .core import ETLPipeline, Job


class CeleryPipeline(ETLPipeline):
    """Celery-based ETL pipeline."""

    def __init__(sel, jobs: list[Job] = None) -> None:
        """Initialize the celery-based ETL pipeline."""
        super().__init__(jobs)

    def _create_chain(self, event, job: Job) -> chain:
        """
        Create a Celery chain associated with a job.

        Parameters
        ----------
        event
            Triggering event.
        job
            Job to associate the task to.

        Returns
        -------
            Celery chain associated with job.
        """
        chained_tasks = chain(
            [job.tasks[0].function.si(event, *job.tasks[0].args, **job.tasks[0].kwargs)]
            + [task.function.s(*task.args, **task.kwargs) for task in job.tasks[1:]]
        )

        return chained_tasks

    def _chain_all_jobs(self, event: Any) -> list[chain]:
        """
        Chain tasks within all jobs.

        Parameters
        ----------
        event
            Triggering event.

        Returns
        -------
            List of Celery chains associated with all jobs.
        """
        if self._jobs is None or len(self._jobs) == 0:
            raise ValueError("No jobs to chain.")
        else:
            return [self._create_chain(event, job) for job in self._jobs]

    def _group_jobs(self, event: Any) -> group:
        """
        Group all jobs by first chaining tasks within all jobs.

        Parameters
        ----------
        event
            Triggering event.

        Returns
        -------
            Celery group associated with all jobs.
        """
        return group(self._chain_all_jobs(event))

    def produce_jobs(self, event: Any) -> Any:
        """
        Produce tasks triggered by an event.

        Parameters
        ----------
        event
            Triggering event.
        """
        group = self._group_jobs(event)
        logging.info("-" * 79)
        logging.info(f"Celery group created: {group}")
        return group.apply_async()



"""dpypeline base package."""
__version__ = "0.1.0-beta.1"
__author__ = "Joao Morado"
__credits__ = "National Oceanography Centre"
__all__ = ["akita", "event_consumer", "etl_pipeline", "celery_app", "filesystems"]



import functools
import logging
import time
from typing import Any, Callable


def retry(
    _func: Callable | None = None, *, max_retries: int = 3, sleep_time: int = 5
) -> Callable:
    """
    Retry decorator for thread-based tasks.

    Parameters
    ----------
    _func
        The function to be decorated.
    max_retries, optional
        Maximum number of retry attempts, by default 3
    sleep_time, optional
        Sleep time between retry attempts, by defaul 5

    Returns
    -------
    decorator
    """

    def decorator_retry(
        func: Callable, max_retries: int = max_retries, sleep_time: int = sleep_time
    ) -> Callable:
        @functools.wraps(func)
        def wrapper_retry(*args, **kwargs) -> Any:
            print(max_retries, sleep_time, args, kwargs)
            n_retries = 0
            while n_retries < max_retries:
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    logging.warning(
                        f"Failed to execute {func.__name__} on attempt {n_retries + 1}."
                    )
                    logging.warning(f"Exception: {e}.")
                    logging.warning(f"Retrying in {sleep_time} seconds.")

                n_retries += 1
                time.sleep(sleep_time)

            raise Exception(
                f"Failed to execute {func.__name__} after {n_retries} attempts."
            )

        return wrapper_retry

    if _func is None:
        # when using @retry(...), the decorated function
        # is not passed as the first argument, and _func is None
        return decorator_retry
    else:
        # when simply using @retry without arguments,
        # the decorated function is passed as the _func argument
        return decorator_retry(_func, max_retries=max_retries, sleep_time=sleep_time)



"""Filehandler for the watchdog."""

import logging
import os
import time
from typing import Any, Protocol

from watchdog.events import (
    DirCreatedEvent,
    DirDeletedEvent,
    DirModifiedEvent,
    DirMovedEvent,
    FileCreatedEvent,
    FileDeletedEvent,
    FileModifiedEvent,
    FileMovedEvent,
    PatternMatchingEventHandler,
)


class Queue(Protocol):
    """Queue interface."""

    def enqueue(self, event: Any) -> Any:
        """Add an event to the queue."""
        ...


class EventHandler(PatternMatchingEventHandler):
    """
    EventHandler for the watchdog.

    Child class of PatternMatchingEventHandler.
    PatternMatchingEventHandler matches given patterns with file paths associated with occurring events.
    Whenever a file in the target path is modified, EventHandler puts the associated event in a queue.
    """

    def __init__(
        self,
        queue: Queue,
        patterns: str | list[str] = None,
        ignore_patterns: str | list[str] | None = None,
        ignore_directories: bool = False,
        case_sensitive: bool = True,
    ) -> None:
        """
        Call the init method of the `PatternMatchingEventHandler` class and sets the queue.

        Parameters
        ----------
        queue
            Queue where events are placed.
        patterns
            Patterns to allow matching events.
        ignore_patterns
            Patterns to ignore matching event paths.
        ignore_directories
            If `True` directories are ignored; `False` otherwise.
        case_sensitive
            If `True` path names are matched sensitive to case; `False` otherwise.
        queue
            Queue where events are placed.
        """
        self._queue = queue

        super().__init__(
            patterns=patterns,
            ignore_patterns=ignore_patterns,
            ignore_directories=ignore_directories,
            case_sensitive=case_sensitive,
        )

    @staticmethod
    def _logging_message(event) -> str:
        """
        Message to log when event is triggered.

        Parameters
        ----------
        event
            Event representing file or directory creation, deletion, modification or moving.

        Returns
        -------
        logging_msg
            Message to be logged.
        """
        logging_msg = f"{event}"
        logging.info("-" * 79)
        logging.info(logging_msg)

        return logging_msg

    @staticmethod
    def is_file_size_stable(path: str, sleep_time: int = 1) -> bool:
        """
        Check if the size of a given file is stable.

        Parameters
        ----------
        path
            Absolute or relative filepath of the file.
        sleep_time
            Sleep time between consecutive file size checks.

        Returns
        -------
        `True` if the file size is stable between two consecutive checks.
        """
        filesize = -1.0

        while filesize != os.path.getsize(path):
            filesize = os.path.getsize(path)
            time.sleep(sleep_time)

        return True

    def _process_event(self, event) -> None:
        """
        Put the event in the queue of events.

        Parameters
        ----------
        event
            Event representing file or directory creation, deletion, modification or moving.
        """
        self._queue.enqueue(event.src_path)

    def on_created(
        self, event: FileCreatedEvent | DirCreatedEvent
    ) -> FileCreatedEvent | DirCreatedEvent:
        """
        Return event when a file or directory is created.

        Parameters
        ----------
        event
            Event representing file or directory creation.

        Returns
        -------
        event
        """
        self._logging_message(event)

        if self.is_file_size_stable(event.src_path):
            self._process_event(event)

        return event

    def on_deleted(
        self, event: FileDeletedEvent | DirDeletedEvent
    ) -> FileDeletedEvent | DirDeletedEvent:
        """
        Return event when a file or directory is deleted.

        Parameters
        ----------
        event
            Event representing file or directory deletion.

        Returns
        -------
        event
        """
        self._logging_message(event)

        return event

    def on_modified(
        self, event: FileModifiedEvent | DirModifiedEvent
    ) -> FileModifiedEvent | DirModifiedEvent:
        """
        Return event when a file or directory is modified.

        Parameters
        ----------
        event
            Event representing file or directory modification.

        Returns
        -------
        event
        """
        self._logging_message(event)

        return event

    def on_moved(
        self, event: FileMovedEvent | DirMovedEvent
    ) -> FileMovedEvent | DirMovedEvent:
        """
        Return event when a file or directory is moved.

        Parameters
        ----------
        event
            Event representing file or directory move/renaming.

        Returns
        -------
        event
        """
        self._logging_message(event)

        return event



"""Akita dependency factory."""

from watchdog.observers.polling import PollingObserver

from .directory_state import DirectoryState
from .event_handler import EventHandler
from .queue_events import EventsQueue


class AkitaFactory:
    """Factory that creates instances required for Akita."""

    def __init__(
        self,
        path: str = "./",
        patterns: str | list[str] = ["*.nc"],
        ignore_patterns: str | list[str] = None,
        ignore_directories: bool = True,
        case_sensitive: bool = True,
        glob_kwargs: dict = None,
    ) -> None:
        """
        Initialize the factory.

        Parameters
        ----------
        path, optional
            _description_, by default "./"
        patterns, optional
            Patterns to allow matching events, by default ["*.nc"]
        ignore_patterns, optional
            Patterns to ignore matching event paths, by default True`
        ignore_directories, optional
            If True directories are ignored, by default True
        case_sensitive, optional
            If True path names are matched sensitive to case, by default True
        glob_kwargs, optional
            Kwargs to pass to glob, by default None
        """
        self._path = path
        self._patterns = list(patterns) if patterns is str else patterns
        self._ignore_patterns = (
            list(ignore_patterns) if ignore_patterns is str else ignore_patterns
        )
        self._ignore_directories = ignore_directories
        self._case_sensitive = case_sensitive
        self._glob_kwargs = glob_kwargs if glob_kwargs is not None else {}

    def get_queue(self, maxsize=0) -> EventsQueue:
        """
        Create the queue.

        Parameters
        ----------
        maxsize
            Maximum size of the queue.

        Returns
        -------
            EventsQueue instance.
        """
        return EventsQueue(maxsize)

    def get_event_handler(
        self,
        queue: EventsQueue,
        patterns: str | list[str] = None,
        ignore_patterns: str | list[str] = None,
        ignore_directories: bool = None,
        case_sensitive: bool = None,
    ) -> EventHandler:
        """
        Create the handler responsible for matching given patterns with file paths associated with occurring events.

        Parameters
        ----------
        patterns
            Patterns to allow matching events.
        ignore_patterns
            Patterns to ignore matching event paths.
        ignore_directories
            If `True` directories are ignored; `False` otherwise.
        case_sensitive
            If `True` path names are matched sensitive to case; `False` otherwise.

        Returns
        -------
            EvetnHandler instance.
        """
        patterns = patterns if patterns is not None else self._patterns
        ignore_patterns = (
            ignore_patterns if ignore_patterns is not None else self._ignore_patterns
        )
        ignore_directories = (
            ignore_directories
            if ignore_directories is not None
            else self._ignore_directories
        )
        case_sensitive = (
            case_sensitive if case_sensitive is not None else self._case_sensitive
        )

        return EventHandler(
            queue, patterns, ignore_patterns, ignore_directories, case_sensitive
        )

    def get_polling_observer(self) -> PollingObserver:
        """
        Create the polling observer.

        Returns
        -------
            PollingObserver instance.
        """
        return PollingObserver()

    def get_directory_state(
        self,
        path: str = None,
        patterns: str | list[str] = None,
        glob_kwargs: dict = None,
    ) -> DirectoryState:
        """
        Create the directory state.

        Parameters
        ----------
        path
            Path to watch.
        patterns, optional
            Patterns to allow matching events.
        glob_kwargs, optional
            Kwargs to pass to glob.

        Returns
        -------
            DirectoryState instance.
        """
        path = path if path is not None else self._path
        patterns = patterns if patterns is not None else self._patterns
        glob_kwargs = glob_kwargs if glob_kwargs is not None else self._glob_kwargs

        return DirectoryState(path, patterns, glob_kwargs)


def get_akita_dependencies(
    path: str = "./",
    patterns: str | list[str] = ["*.nc"],
    ignore_patterns: str | list[str] = None,
    ignore_directories: bool = True,
    case_sensitive: bool = True,
    glob_kwargs: dict = None,
) -> tuple[str, EventsQueue, EventHandler, DirectoryState, PollingObserver]:
    """
    Get the dependencies of Akita.

    Returns
    -------
        List of dependencies of Akita.
    """
    akita_factory = AkitaFactory(
        path, patterns, ignore_patterns, ignore_directories, case_sensitive, glob_kwargs
    )
    queue = akita_factory.get_queue()
    event_handler = akita_factory.get_event_handler(queue)
    directory_state = akita_factory.get_directory_state()
    polling_observer = akita_factory.get_polling_observer()

    return path, queue, event_handler, directory_state, polling_observer



"""Directory state class."""
import glob
import logging
import os
import pickle


class DirectoryState:
    """DirectoryState class."""

    _state_file: str = os.path.join(os.getenv("CACHE_DIR"), "directory_state.pickle")

    def __init__(
        self, path: str, patterns: str | list[str], glob_kwargs: dict = None
    ) -> None:
        """
        Initiate the DirectoryState class.

        Parameters
        ----------
        path
            Path to watch.
        patterns
            Patterns for matching files.
        """
        self._current_state: list[str] = None
        self._stored_state: list[str] = None

        self._path = path
        self._patterns = patterns if patterns else list(patterns)
        self._glob_kwargs = glob_kwargs if glob_kwargs is not None else {}

        assert (
            os.getenv("CACHE_DIR") is not None
        ), "CACHE_DIR environmental variable is not set."

    @property
    def stored_state(self) -> list[str]:
        """
        Return the stored state of the directory.

        Returns
        -------
            Store state of the directory.
        """
        self._load_state()
        return self._stored_state

    @property
    def current_state(self) -> list[str]:
        """
        Return the current state of the directory.

        Returns
        -------
            Current state of the directory.
        """
        self._get_current_directory_state()
        return self._current_state

    def _save_state(self) -> None:
        """Save the state of the directory by only saving files that were already enqueued."""
        logging.info("-" * 79)
        logging.info("Saving state of the directory.")

        with open(self._state_file, "wb") as f:
            pickle.dump(self._current_state, f)

    def _load_state(self) -> None:
        """Load the stored state of the directory."""
        logging.info("-" * 79)
        logging.info("Loading state of the directory.")

        if os.path.isfile(self._state_file):
            logging.info(f"Found stored directory state file {self._state_file}.")
            with open(self._state_file, "rb") as f:
                self._stored_state = pickle.load(f)
        else:
            logging.info(f"No directory state file {self._state_file} was found.")
            self._stored_state = []

    def _get_current_directory_state(self) -> list[str]:
        """
        Get and set the current state of the directory.

        Notes
        -----
        Everytime this method is called, the current state of the directory is saved.

        Returns
        -------
            Current state of the directory.
        """
        self._current_state = []
        for pattern in self._patterns:
            found_files = glob.glob(self._path + pattern, **self._glob_kwargs)
            self._current_state.extend(found_files)

        self._save_state()

        return self._current_state



"""Queue of events class."""
from __future__ import annotations

import logging
import os
import pickle
from queue import Empty, Full, Queue
from typing import Any


class EventsQueue(Queue):
    """EventsQueue singleton class."""

    _instance: EventsQueue = None
    _initialized: bool = False
    _state_file: str = os.path.join(os.getenv("CACHE_DIR"), "queue_state.pickle")
    _processed_events_file: str = os.path.join(
        os.getenv("CACHE_DIR"), "processed_events.pickle"
    )

    def __init__(self, maxsize: int = 0) -> None:
        """Initiate the EventsQueue singleton."""
        if not self._initialized:
            logging.info("Initializing singleton instance of EventsQueue.")
            super().__init__(maxsize=maxsize)
            self._initialized: bool = True
            self._processed_events: list[str] = None
            self._load_state()

        assert (
            os.getenv("CACHE_DIR") is not None
        ), "CACHE_DIR environmental variable is not set."

    def __new__(cls, maxsize: int = 0) -> EventsQueue:
        """
        Create a new instance of the EventsQueue class using the Singleton pattern.

        Returns
        -------
            Instance of the EventsQueue class.
        """
        if cls._instance is None:
            logging.info("-" * 79)
            logging.info("Creating singleton instance of EventsQueue.")
            cls._instance = super(EventsQueue, cls).__new__(cls)

        return cls._instance

    @property
    def queue_list(self) -> list[Any]:
        """
        Return the list of events in the queue.

        Returns
        -------
            List of events in the queue.
        """
        return list(self.queue)

    @property
    def processed_events(self) -> list[Any]:
        """
        Return the list of events processed thus far.

        Returns
        -------
            List of events processed thus far.
        """
        if self._processed_events is None:
            # Only load processed events if first time called.
            self._load_processed_events()

        return self._processed_events

    def _save_state(self, logging_prefix="") -> None:
        """Save the state of the queue."""
        logging.info("-" * 79)
        logging.info(f"{logging_prefix}Saving state of the queue.")

        with open(self._state_file, "wb") as f:
            pickle.dump(self.queue, f)

    def _load_state(self) -> None:
        """Load the state of the queue."""
        logging.info("-" * 79)
        logging.info("Loading state of the queue.")

        if os.path.isfile(self._state_file):
            logging.info(f"Found queue state file {self._state_file}")

            with open(self._state_file, "rb") as f:
                self.queue = pickle.load(f)
        else:
            logging.info(f"No queue state file {self._state_file} was found.")

    def _load_processed_events(self) -> None:
        """Load the processed events."""
        logging.info("-" * 79)
        logging.info("Loading processed events.")

        if os.path.isfile(self._processed_events_file):
            logging.info(f"Found processed events file {self._processed_events_file}.")

            with open(self._processed_events_file, "rb") as f:
                self._processed_events = pickle.load(f)
        else:
            logging.info(
                f"No processed events file {self._processed_events_file} was found."
            )
            self._processed_events = []

    def _save_processed_events(self) -> None:
        """Save the events processed so far in this session."""
        logging.info("-" * 79)
        logging.info("Saving processed events.")

        with open(self._processed_events_file, "wb") as f:
            pickle.dump(self._processed_events, f)

    def enqueue(self, event) -> bool:
        """Add an event to the queue.

        Everytime an event is added to the queue, the state of the queue is saved.

        Parameters
        ----------
        event
            Event to add to the queue.

        Returns
        -------
        True if the event was added to the queue.
        """
        try:
            logging.info("-" * 79)
            logging.info(f"Enqueuing event: {event}.")
            self.put(event, block=False)
            self._save_state(logging_prefix="Enqueueing event: ")
            return True
        except Full:
            raise Full("Queue is full.")

    def dequeue(self) -> Any:
        """
        Remove and return an event from the queue.

        If the queue is empty, returns None.

        Returns
        -------
        First event in the queue or None if the queue is empty.
        """
        try:
            logging.info("-" * 79)
            logging.info("Dequeuing event: Starting.")
            event = self.get(block=False)
            logging.info(f"Dequeued event: {event}.")
            self._save_state(logging_prefix="Dequeuing event: ")

            # Set as processesed event
            if self._processed_events is None:
                self._processed_events = []

            self._processed_events.append(event)
            self._save_processed_events()
            return event
        except Empty:
            return None

    def peek(self) -> Any:
        """
        Peek first item in the queue withou removing it.

        If the queue is empty, returns None.

        Returns
        -------
        First event in the queue or None if the queue is empty.
        """
        if self.get_queue_size():
            logging.info("-" * 79)
            logging.info("Peeking first item in the queue.")
            return self.queue[0]

    def get_queue_size(self) -> int:
        """
        Get the size of the queue.

        Returns
        -------
        Number of events in the queue.
        """
        return self.qsize()

    @classmethod
    def clear_instance(cls):
        """Clear the singleton instance."""
        cls._instance = None



"""Event consumer. Acts as an interface between Akita and the ETL pipeline."""
import logging
import time
from threading import Thread
from typing import Any, Protocol


class JobProducer(Protocol):
    """Job producer interface."""

    def produce_jobs(self, event: Any) -> None:
        """Produce jobs triggered by an event."""
        ...


class Queue(Protocol):
    """Queue interface."""

    def dequeue(self) -> Any:
        """Remove and return an event from the queue."""
        ...

    def peek(self) -> Any:
        """Peek first item in the queue withou removing it."""
        ...

    def get_queue_size(self) -> int:
        """Return the size of the queue."""
        ...


class EventConsumer:
    """
    Event consumer that runs on a thread as a daemon process.

    This event consumer consumes events from an in-memory queue and processes them to produce jobs.

    Attributes
    ----------
    _queue
        Queue where events are placed.
    _job_producer
        Producer of jobs.
    _worker, optional
        Worker thread that consumes events from the and processes them to produce jobs.
    """

    def __init__(
        self, queue: Queue, job_producer: JobProducer, worker: Thread = None
    ) -> None:
        """
        Initialize the event consumer.

        Parameters
        ----------
        queue
            Queue where events are placed.
        job_producer
            Producer of jobs.
        worker, optional
            Worker thread that consumes events from the and processes them to produce jobs.
        """
        self._job_producer = job_producer
        self._queue = queue
        self._worker = worker

    def _consume_event(self, event) -> None:
        """
        Consume event by producing jobs.

        Parameters
        ----------
        event
            Event representing file or directory creation, deletion, modification or moving.
        """
        self._job_producer.produce_jobs(event=event)

    def _run_worker(self, sleep_time: int = 1) -> None:
        """
        Run the worker thread.

        Callback function to be used as a target by Thread.

        Parameters
        ----------
        sleep_time
            Sleep time in seconds for which the thread is idle.
        """
        while True:
            if self._queue.get_queue_size():
                event = self._queue.peek()
                logging.info("-" * 79)
                logging.info(f"Consuming event: {event}")
                self._consume_event(event)
                logging.info(f"Event consumed: {event}")
                self._queue.dequeue()
            else:
                time.sleep(sleep_time)

    def _create_worker(self, daemon: bool = True) -> Thread:
        """
        Create the worker thread.

        Parameters
        ----------
        daemon
            If `True`, runs the thread as a daemon; otherwise thread is not created as a daemon.

        Returns
        -------
            Worker thread that consumes events from the in-memory queue and processes them to produce jobs.
        """
        # Set up a worker thread to process database load
        self._worker = Thread(target=self._run_worker, daemon=daemon)

        return self._worker

    def run(self) -> None:
        """Run the event consumer."""
        try:
            if self._worker is None:
                self._create_worker()

            self._worker.start()
        except Exception:
            self._worker.join()



"""ETL pipeline definitions."""
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Any


@dataclass(frozen=True)
class Task:
    """
    Task class.

    Attributes
    ----------
    function
        Task function.
    args
        Arguments to be passed to the function.
    kwargs
        Keyword arguments to be passed to the function.
    """

    function: Any
    args: tuple = field(default_factory=tuple)
    kwargs: dict = field(default_factory=dict)


@dataclass
class Job:
    """
    A job is a collection of tasks.

    Parameters
    ----------
    name
        Job name.
    tasks
        List of tasks.
    """

    name: str = field(default_factory=str)
    tasks: list[Task] = field(default_factory=list)

    def add_task(self, task: Task) -> None:
        """
        Add a task to the job.

        Parameters
        ----------
        task
            Task to add.
        """
        self.tasks.append(task)

    def remove_task(self, index: int, task: Task = None) -> Task:
        """
        Remove a task from the job.

        Parameters
        ----------
        index
            Index of the task to remove.
        task, optional
            Task to remove.

        Returns
        -------
            Removed task.
        """
        if task is Task:
            return self.tasks.pop(self.tasks.index(task))
        else:
            return self.tasks.pop(index)


class ETLPipeline(ABC):
    """Abstract base class for ETL pipelines."""

    def __init__(self, jobs: list[Job] = None) -> None:
        """Initialize the ETL pipeline."""
        self._jobs = jobs if jobs is not None else []

    @property
    def jobs(self) -> list[Job]:
        """List of jobs."""
        return self._jobs

    def add_job(self, job: Job) -> None:
        """
        Add a job to the pipeline.

        Parameters
        ----------
        job
            Job to add.
        """
        return self._jobs.append(job)

    def remove_job(self, index: int, job: Job = None) -> Any:
        """
        Remove a job from the pipeline.

        Parameters
        ----------
        index
            Index of the job to remove.
        job, optional
            Job to remove.

        Returns
        -------
            Removed job.
        """
        if job is Job:
            return self._jobs.pop(self._jobs.index(job))
        else:
            return self._jobs.pop(index)

    @abstractmethod
    def produce_jobs(self, event: Any) -> Any:
        """Abstraction for producing tasks triggered by an event."""
        pass



"""Akita, the watchdog class."""
import logging
import time
from itertools import chain
from typing import Any, Protocol


class Queue(Protocol):
    """Queue class protocol."""

    def enqueue(self, event: Any) -> bool:
        """Enqueue an event."""
        ...

    @property
    def queue_list(self) -> list[Any]:
        """Return the list of events in the queue."""
        ...

    @property
    def processed_events(self) -> list[Any]:
        """Return the list of processed events."""
        ...


class DirectoryState(Protocol):
    """DirectoryState class protocol."""

    @property
    def stored_state(self) -> list[str]:
        """Return the stored state of the directory."""
        ...

    @property
    def current_state(self) -> list[str]:
        """Return the current state of the directory."""


class EventHandler(Protocol):
    """EventHandler class protocol."""

    def on_created(self, event: Any) -> None:
        """Return event when a file or directory is created."""
        ...

    def on_deleted(self, event: Any) -> None:
        """Return event when a file or directory is deleted."""
        ...

    def on_modified(self, event: Any) -> None:
        """Return event when a file or directory is modified."""
        ...

    def on_moved(self, event: Any) -> None:
        """Return event when a file or directory is moved."""
        ...


class Observer(Protocol):
    """Observer class protocol."""

    def start(self) -> None:
        """Start the obsever."""
        ...

    def stop(self) -> None:
        """Stop the observer."""
        ...

    def join(self) -> None:
        """Join the observer."""
        ...

    def schedule(self, event_handler: EventHandler, path: str, recursive: bool) -> None:
        """Schedules watching a path."""
        ...


class Akita:
    """
    Akita, the watchdog.

    The Akita watchdog holds the queue, observer, and event handler instances.
    The in-memory Queue is created upon instantiation and should be a singleton to provide a global point of access.

    Attributes
    ----------
    _path
        Path to watch.
    _queue
        Queue where events are placed by the event handler.
    _event_handler
        Handler responsible for matching given patterns with file paths associated with occurring events.
    _directory_state
        Instance that holds the state of the directory.
    _observer
        Observer thread that schedules watching directories and dispatches calls to event handlers.
    _event_handler
        Handler responsible for matching given patterns with file paths associated with occurring events.
    """

    def __init__(
        self,
        path: str,
        queue: Queue,
        event_handler: EventHandler,
        directory_state: DirectoryState,
        observer: Observer,
        run_init: bool = False,
    ) -> None:
        """
        Initialize the Akita watchdog.

        Parameters
        ----------
        path
            Path to watch.
        queue
            Queue where events are placed by the event handler.
        event_handler
            Handler responsible for matching given patterns with file paths associated with occurring events.
        directory_state
            Instance that holds the state of the directory.
        observer
            Observer thread that schedules watching directories and dispatches calls to event handlers.
        run_init
            If `True` runs the watchdog. `False` otherwise.
        """
        self._path = path
        self._queue = queue
        self._observer = observer
        self._event_handler = event_handler
        self._directory_state = directory_state

        if run_init:
            self.run()

    @property
    def queue(self):
        """Return the queue."""
        return self._queue

    @queue.deleter
    def queue(self):
        """Delete the queue."""
        logging.info("-" * 79)
        logging.info("Deleting in-memory queue.")
        logging.info("-" * 79)

        del self._queue

    def _get_unenqueued_files(self) -> list[str]:
        """
        Return a list of files that are not enqueued.

        Notes
        -----
        At the beggining of the run,
          Akita determines which files in the the directory being monitored have never been enqueued.
        This is done as follows:

        n = d - (s V q V p)

        where

        - n: events not enqueued.
        - s: events in the stored directory state.
        - d: events in the current directory state.
        - q: events in the queue when the previous session terminated.
        - p: events processed in the previous session.

        Returns
        -------
            List of files to enqueue.
        """
        s = self._directory_state.stored_state
        d = self._directory_state.current_state
        q = self._queue.queue_list
        p = self._queue.processed_events

        n = set(d) - set(chain.from_iterable([s, q, p]))

        logging.info("-" * 79)
        logging.info(f"Events in the stored directory state (n={len(s)}):")
        for event in s:
            logging.info(f"{event}")

        logging.info(f"Events in the current directory state (n={len(d)}):")
        for event in d:
            logging.info(f"{event}")

        logging.info(
            f"Events in the queue when the previous session terminated (n={len(q)}):"
        )
        for event in q:
            logging.info(f"{event}")

        logging.info(
            f"Events processed before the previous session terminated (n={len(p)}):"
        )
        for event in p:
            logging.info(f"{event}")

        logging.info(f"Events unenqueued (n={len(n)}):")
        for event in n:
            logging.info(f"{event}")

        logging.info(
            f"Found {len(n)} files not enqueued in the current state of the directory."
        )

        logging.info("-" * 79)

        return list(n)

    def _enqueue_new_files(self) -> None:
        """Enqueue events previously unqueued."""
        # Get current state of the directory
        not_enqueued_state = sorted(self._get_unenqueued_files())

        for event in not_enqueued_state:
            self._queue.enqueue(event)

    def run(self) -> None:
        """Run the Akita watchdog."""
        # Add new files found in the directory since the last time the watchdog was run.
        self._enqueue_new_files()

        logging.info("Starting the Akita watchdog.")
        self._observer.schedule(self._event_handler, self._path, recursive=True)
        self._observer.start()

        try:
            while True:
                time.sleep(1)
        finally:
            self._observer.stop()

        self._observer.join()
